#include "definitions.h"

typedef std::complex<double> complex;

//  This problem describes the distribution of the vector potential in
//  a 2D domain comprising a wire carrying electrical current, air, and
//  an iron which is not under voltage.
//
//  PDE: -(1/mu)Laplace A + ii*omega*gamma*A - J_ext = 0.
//
//  Domain: Rectangle of height 0.003 and width 0.004. Different
//  materials for the wire, air, and iron (see mesh file domain2.mesh).
//
//  BC: Zero Dirichlet on the top and left and right edges, zero Neumann
//  on the bottom.
//
//  The following parameters can be changed:
// Number of initial uniform mesh refinements.
const int INIT_REF_NUM = 0;
// Initial polynomial degree of all mesh elements.
const int P_INIT = 1;
// This is a quantitative parameter of Adaptivity.
const double THRESHOLD = 0.9;

// Error calculation & adaptivity.
DefaultErrorCalculator<::complex, HERMES_H1_NORM> errorCalculator(RelativeErrorToGlobalNorm, 1);
// Stopping criterion for an adaptivity step.
AdaptStoppingCriterionSingleElement<::complex> stoppingCriterion(THRESHOLD);
// Adaptivity processor class.
Adapt<::complex> adaptivity(&errorCalculator, &stoppingCriterion);
// Predefined list of element refinement candidates.
const CandList CAND_LIST = H2D_HP_ANISO;
// Stopping criterion for adaptivity.
const double TOTAL_ERROR_ESTIMATE_STOP = 1e-3;

// Problem parameters.
const double MU_0 = 4.0*M_PI*1e-7;
const double MU_IRON = 1e3 * MU_0;
const double GAMMA_IRON = 6e6;
const double J_EXT = 1e6;
const double FREQ = 5e3;
const double OMEGA = 2 * M_PI * FREQ;

int main(int argc, char* argv[])
{
  Hermes::Mixins::TimeMeasurable m;
  m.tick();

  // Load the mesh.
  MeshSharedPtr mesh(new Mesh);
  MeshReaderH2D mloader;
  mloader.load("domain.mesh", mesh);

  // Perform initial mesh refinements.
  for (int i = 0; i < INIT_REF_NUM; i++) mesh->refine_all_elements();

  // Initialize boundary conditions.
  Hermes::Hermes2D::DefaultEssentialBCConst<::complex> bc_essential("Dirichlet",::complex(0.0, 0.0));
  EssentialBCs<::complex> bcs(&bc_essential);

  // Create an H1 space with default shapeset.
  SpaceSharedPtr<::complex> space(new H1Space<::complex>(mesh, &bcs, P_INIT));
  int ndof = space->get_num_dofs();

  // Initialize the weak formulation.
  WeakFormSharedPtr<::complex> wf(new CustomWeakForm("Air", MU_0, "Iron", MU_IRON, GAMMA_IRON,
    "Wire", MU_0,::complex(J_EXT, 0.0), OMEGA));

  // Initialize coarse and reference mesh solution.
  MeshFunctionSharedPtr<::complex> sln(new Hermes::Hermes2D::Solution<::complex>());
  MeshFunctionSharedPtr<::complex> ref_sln(new Hermes::Hermes2D::Solution<::complex>());

  // Initialize refinement selector.
  H1ProjBasedSelector<::complex> selector(CAND_LIST);

  // Initialize views.
  Views::ScalarView sview("Solution (real part)", new Views::WinGeom(0, 0, 600, 350));
  Views::ScalarView sview_error("Error function", new Views::WinGeom(0, 360, 600, 350));
  Views::OrderView oview("Polynomial orders", new Views::WinGeom(610, 0, 520, 350));

  // DOF and CPU convergence graphs initialization.
  SimpleGraph graph_dof, graph_cpu;

  DiscreteProblem<::complex> dp(wf, space);

  // Perform Newton's iteration and translate the resulting coefficient vector into a Solution.
  Hermes::Hermes2D::NewtonSolver<::complex> newton(&dp);
  SimpleGraph graph_dof_est;

  // Adaptivity loop:
  int as = 1; bool done = false;
  adaptivity.set_space(space);
  do
  {
    // Construct globally refined reference mesh and setup reference space.
    Mesh::ReferenceMeshCreator ref_mesh_creator(mesh);
    MeshSharedPtr ref_mesh = ref_mesh_creator.create_ref_mesh();
    Space<::complex>::ReferenceSpaceCreator ref_space_creator(space, ref_mesh);
    SpaceSharedPtr<::complex> ref_space = ref_space_creator.create_ref_space();

    newton.set_space(ref_space);

    int ndof_ref = ref_space->get_num_dofs();

    // Initialize reference problem.

    // Initial coefficient vector for the Newton's method.
   ::complex* coeff_vec = new::complex[ndof_ref];
    memset(coeff_vec, 0, ndof_ref * sizeof(::complex));

    // Perform Newton's iteration and translate the resulting coefficient vector into a Solution.
    newton.solve(coeff_vec);
    Hermes::Hermes2D::Solution<::complex>::vector_to_solution(newton.get_sln_vector(), ref_space, ref_sln);

    // Project the fine mesh solution onto the coarse mesh.
    OGProjection<::complex> ogProjection;
    ogProjection.project_global(space, ref_sln, sln);

    // View the coarse mesh solution and polynomial orders.
    MeshFunctionSharedPtr<double> real_filter(new RealFilter(sln));
    MeshFunctionSharedPtr<double> rreal_filter(new RealFilter(ref_sln));
    sview.show(rreal_filter);
    sview.save_numbered_screenshot("sln%02d.bmp", as, true);

    oview.show(ref_space);
    oview.save_numbered_screenshot("refSpace%02d.bmp", as, true);

    // Calculate element errors and total error estimate.
    errorCalculator.calculate_errors(sln, ref_sln);
    std::cout << (std::string)"Relative error: " << errorCalculator.get_total_error_squared() * 100. << '%' << std::endl;

    // Add entry to DOF and CPU convergence graphs.
    graph_dof_est.add_values(space->get_num_dofs(), errorCalculator.get_total_error_squared() * 100.);
    sview_error.show(errorCalculator.get_errorMeshFunction());
    sview_error.save_numbered_screenshot("errorView%02d.bmp", as, true);

    // If err_est too large, adapt the mesh.
    if (errorCalculator.get_total_error_squared()  * 100. < TOTAL_ERROR_ESTIMATE_STOP)
      done = true;
    else
    {
      std::cout << (std::string)"Adapting..." << std::endl << std::endl;
      adaptivity.adapt(&selector);
    }

    // Clean up.
    delete[] coeff_vec;

    // Increase counter.
    as++;
  } while (done == false);

  graph_dof_est.save("conv_dof_est.dat");

  // Show the reference solution - the final result.
  sview.set_title("Fine mesh solution");

  MeshFunctionSharedPtr<double> real_filter(new RealFilter(ref_sln));
  sview.show(real_filter);

  m.tick();
  std::cout << m.accumulated();

  // Wait for all views to be closed.
  Views::View::wait();
  return 0;
}